package ru.ifmo.hmm;

import static java.lang.Math.*;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

public class Hmm {

    final double a[][];
    final double b[][];
    final double h[];

    // distributions for sampling
    private final double da[][];
    private final double db[][];
    private final double dh[];

    // for 3d problem (updating)
    private double na[][][];
    private double nb[][][];
    private double nh[];

    // aux
    final int n;
    final int m;
    private final Random rnd = new Random(DataProvider.SEED);
    private static final boolean CHECK_MODEL = false;
    private static final boolean UPDATE_SAMPLING = false;

    public Hmm(int n, int m) {
        this(new double[n][n], new double[n][m], new double[n], true);
    }

    public Hmm(double[][] a, double[][] b, double[] p) {
        this(a, b, p, false);
    }

    private Hmm(double[][] a, double[][] b, double[] p, boolean random) {
        n = a.length;
        m = b[0].length;

        this.a = a;
        this.b = b;
        this.h = p;

        // for sampling
        da = new double[n][n];
        db = new double[n][m];
        dh = new double[n];

        if (random) {
            rndUniform();
        } else {
            updateSampling();
        }

        checkModel();

        // for 3d problem
        na = new double[n][n][2];
        nb = new double[n][m][2];
        nh = new double[n];
    }

    public void rndUniform() {
        rndUniform(a);
        rndUniform(b);
        rndUniform(h);
        updateSampling();
        checkModel();
    }

    public void randomizeTopology(double delta) {
        randomize(delta);
        rndTop(a, 0.95);
        rndTop(b, 0.2);
        rndTop(h, 0.2);
        updateSampling();
        checkModel();
    }

    private void rndTop(double[] x, double pr) {
        int nonZero = 0;
        for (double xx : x) {
            if (abs(xx) > 1e-10) nonZero++;
        }
        for (int i = 0; i < x.length && nonZero > 1; i++) {
            if (rnd.nextDouble() < pr) {
                x[i] = 0;
                nonZero--;
            }
        }

        normalize(x);
    }

    private void rndTop(double[][] x, double pr) {
        for (double[] xx : x) {
            rndTop(xx, pr);
        }
    }

    public void randomizeLeftRight(double delta) {
        randomize(delta);
        makeLeftRight(a);
    }

    public void randomize(double delta) {
        for (double[] aa : a) {
            randomize3(aa, delta);
        }

        for (double[] bb : b) {
            randomize3(bb, delta);
        }

        randomize3(h, delta);

        updateSampling();
        checkModel();
    }

    private void makeLeftRight(double[][] a) {
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < i; j++) {
                a[i][j] = 0;
            }
        }
        normalize(a);
    }

    private void normalize(double[][] a) {
        for (double[] aa : a) {
            normalize(aa);
        }
    }

    private void randomize2(double[] x, double delta) {
        double m = 1.0 / x.length;
        double sum = 0;
        delta = min(delta, m / (n - 1));
        for (int i = 0; i < x.length - 1; i++) {
            x[i] = m - delta + rnd.nextDouble() * (2 * delta);
            sum += x[i];
        }
        x[x.length - 1] = 1 - sum;
    }

    private void randomize3(double[] x, double delta) {
        for (int i = 0; i < x.length; i++) {
            x[i] = abs(rnd.nextGaussian()) /** delta*/;
        }
        normalize(x);
    }

    private void normalize(double[] x) {
        double sum = 0;
        for (double xx : x) {
            sum += xx;
        }
        for (int i = 0; i < x.length; i++) {
            x[i] /= sum;
        }
    }

    private void randomize(double[] x, double delta) {
        double m = 1.0 / x.length;
        double sum = 0;
        delta = min(delta, m / (n - 1));
        for (int i = 0; i < x.length - 1; i++) {
            double d = rnd.nextGaussian() * delta;
            d = min(d, 1 - sum - m);
            d = max(d, -m);
            sum += x[i] = m + d;
        }
        x[x.length - 1] = 1 - sum;
    }

    private void checkModel() {
        if (!CHECK_MODEL) return;
        checkNull("A", a);
        checkNull("B", b);
        checkNull("P", h);
        checkDistr("A", a);
        checkDistr("B", b);
        checkDistr("P", h);
    }

    private void checkDistr(String s, double[][] x) {
        for (int i = 0; i < x.length; i++) {
            checkDistr(s + "[" + i + "]", x[i]);
        }
    }

    private void checkDistr(String s, double[] x) {
        double sum = 0;
        for (int i = 0; i < x.length; i++) {
            if (x[i] < 0) throw new IllegalArgumentException(s + "[" + i + "] = " + x[i] + " < 0");
            sum += x[i];
        }
        if (abs(1 - sum) > 1e-10) {
            throw new IllegalArgumentException(s + " has sum: " + sum + " != 1");
        }
    }

    private void checkNull(String s, Object x) {
        if (x == null) {
            throw new IllegalArgumentException(s + " == null");
        }
    }

    // initial params estimation
    public void estimate(int ss[][], double epsLn) {
        double p1, p2 = solve1Ln(ss);
        do {
            p1 = p2;
            nextEstimate(ss);
            p2 = solve1Ln(ss);
        } while (abs(p1 - p2) > epsLn);
        updateSampling();
        checkModel();
    }

    private void nextEstimate(int[][] ss) {
        int[][] pb = new int[n][m];

        int[] start = new int[n];
        int[][] pa = new int[n][n];
        int cnt[] = new int[n];
        for (int[] s : ss) {
            int[] path = solve2b(s, cnt);
            start[path[0]]++;
            for (int i = 0; i < s.length; i++) {
                pb[path[i]][s[i]]++;
                if (i < s.length - 1) {
                    pa[path[i]][path[i + 1]]++;
                }
            }
        }

        for (int i = 0; i < n; i++) {
            h[i] = start[i] * 1.0 / ss.length;

            int allB = 0;
            for (int j = 0; j < m; j++) {
                allB += pb[i][j];
            }

            if (allB > 0) {
                for (int j = 0; j < m; j++) {
                    b[i][j] = pb[i][j] * 1.0 / allB;
                }
            }

            int allA = 0;
            for (int j = 0; j < n; j++) {
                allA += pa[i][j];
            }
            if (allA > 0) {
                for (int j = 0; j < n; j++) {
                    a[i][j] = pa[i][j] * 1.0 / allA;
                }
            }
        }


    }

    // seq generator
    public int[] generate(int k) {
        int s = sampler(dh);
        int res[] = new int[k];
        for (int i = 0; i < k; i++) {
            res[i] = sampler(db[s]);
            s = sampler(da[s]);
        }
        return res;
    }

    public double solve1(int ss[][]) {
        return exp(solve1Ln(ss));
    }

    public double solve1Ln(int ss[][]) {
        double p = 0;
        for (int[] s : ss) {
            double z = solve1(s);
            p += log(z);
        }
        return p;
    }

    public double solve1(int s[]) {
        double[][] pa = getA(s);
        return sum(pa[s.length - 1]);
    }

    // ML t-th 'bit'
    public int[] solve2a(int s[]) {
        double[][] pa = getA(s);
        double[][] pb = getB(s);
        int r[] = new int[s.length];
        for (int t = 0; t < r.length; t++) {
            int ind = 0;
            for (int i = 1; i < n; i++) {
                if (pa[t][i] * pb[t][i] > pa[t][ind] * pb[t][ind]) {
                    ind = i;
                }
            }
            r[t] = ind;
        }
        return r;
    }

    // ML sequence
    public int[] solve2b(int s[], int[] cnt) {
        double p[][] = new double[s.length][n];
        for (int i = 0; i < n; i++) {
            p[0][i] = h[i] * b[i][s[0]];
        }
        for (int t = 1; t < s.length; t++) {
            for (int i = 0; i < n; i++) {
                double pr = 0;
                for (int j = 0; j < n; j++) {
                    pr = max(p[t - 1][j] * a[j][i], pr);
                }
                p[t][i] = pr * b[i][s[t]];
            }
        }

        int v = 0;
        for (int i = 1; i < n; i++) {
            double v1 = p[s.length - 1][i];
            double v2 = p[s.length - 1][v];
            if (v1 > v2 || v1 == v2 && cnt[i] < cnt[v]) {
                v = i;
            }
        }

        int res[] = new int[s.length];
        res[s.length - 1] = v;
        cnt[v]++;
        for (int t = s.length - 2; t >= 0; t--) {
            int u = 0;
            for (int i = 1; i < n; i++) {
                if (p[t][i] > p[t][u] || p[t][i] == p[t][u] && cnt[i] < cnt[u]) {
                    u = i;
                }
            }
            res[t] = v = u;
            cnt[v]++;
        }
        return res;
    }

    public double solve3(int ss[][], double epsLn) {
        double p1, p2 = solve1Ln(ss);
        do {
            p1 = p2;
            nextBW(ss);
            p2 = solve1Ln(ss);
        } while (abs(p2 - p1) > epsLn);
        updateSampling();
        checkModel();
        return solve1Ln(ss);
    }

    private void dist(double y[], double x[]) {
        Arrays.fill(y, 0);
        for (int i = 1; i < x.length; i++) {
            y[i] = y[i - 1] + x[i - 1];
        }
    }

    private void dist(double y[][], double[][] x) {
        for (int i = 0; i < x.length; i++) {
            dist(y[i], x[i]);
        }
    }

    private int sampler(double[] s) {
        int i = Arrays.binarySearch(s, rnd.nextDouble());
        return i >= 0 ? i : -(i + 1) - 1;
    }

    public double[][] getA(int s[]) {
        double pa[][] = new double[s.length][n];
        for (int i = 0; i < n; i++) {
            pa[0][i] = h[i] * b[i][s[0]];
        }
        for (int t = 1; t < s.length; t++) {
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < n; j++) {
                    pa[t][i] += pa[t - 1][j] * a[j][i];
                }
                pa[t][i] *= b[i][s[t]];
            }
        }
        return pa;
    }

    public double[][] getB(int s[]) {
        double pb[][] = new double[s.length][n];
        Arrays.fill(pb[s.length - 1], 1);
        for (int t = s.length - 2; t >= 0; t--) {
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < n; j++) {
                    pb[t][i] += a[i][j] * b[j][s[t + 1]] * pb[t + 1][j];
                }
            }
        }
        return pb;
    }

    private void nextBW(int ss[][]) {
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                na[i][j][0] = na[i][j][1] = 0;
            }
        }
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < m; j++) {
                nb[i][j][0] = nb[i][j][1] = 0;
            }
        }
        Arrays.fill(nh, 0);

        int skipped = 0;
        for (int[] s : ss) {
            double[][] pa = getA(s);
            double[][] pb = getB(s);
            double p = solve1(s);
            if (p == 0) {
                skipped++;
                continue;
            }
            for (int i = 0; i < n; i++) {

                // exp num of moves from state i
                double c2 = 0;
                for (int t = 0; t < s.length - 1; t++) {
                    c2 += pa[t][i] * pb[t][i];
                }

                // update a
                for (int j = 0; j < n; j++) {
                    double c1 = 0;
                    for (int t = 0; t < s.length - 1; t++) {
                        c1 += pa[t][i] * a[i][j] * b[j][s[t + 1]] * pb[t + 1][j];
                    }
                    na[i][j][0] += c1 / p;
                    na[i][j][1] += c2 / p;
                }

                // exp num of times in state i
                c2 += pa[s.length - 1][i] * pb[s.length - 1][i];

                // update b
                for (int j = 0; j < m; j++) {
                    double c1 = 0;
                    for (int t = 0; t < s.length; t++) {
                        if (s[t] == j) {
                            c1 += pa[t][i] * pb[t][i];
                        }
                    }
                    nb[i][j][0] += c1 / p;
                    nb[i][j][1] += c2 / p;
                }

                nh[i] += pa[0][i] * pb[0][i] / p;
            }
        }

        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                if (na[i][j][1] != 0) {
                    a[i][j] = na[i][j][0] / na[i][j][1];
                }
            }
        }
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < m; j++) {
                if (nb[i][j][1] != 0) {
                    b[i][j] = nb[i][j][0] / nb[i][j][1];
                }
            }
        }
        if (ss.length > skipped) {
            for (int i = 0; i < n; i++) {
                h[i] = nh[i] / (ss.length - skipped);
            }
        }
    }

    public double sum(double[] x) {
        double r = 0;
        for (double v : x) {
            r += v;
        }
        return r;
    }

    public void updateSampling() {
        updateSampling(false);
    }

    public void updateSampling(boolean force) {
        if (!force && !UPDATE_SAMPLING) {
            return;
        }
        dist(da, a);
        dist(db, b);
        dist(dh, h);
    }

    private void rndUniform(double[] x) {
        Arrays.fill(x, 1.0 / x.length);
    }

    private void rndUniform(double[][] a) {
        for (double[] aa : a) {
            Arrays.fill(aa, 1.0 / aa.length);
        }
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (double[] aa : a) {
            for (int j = 0; j < n; j++) {
                sb.append(String.format("%.3f ", aa[j]));
            }
            sb.append("\n");
        }
        sb.append("\n");

        for (int i = 0; i < n; i++) {
            for (int j = 0; j < m; j++) {
                sb.append(String.format("%.3f ", b[i][j]));
            }
            sb.append("\n");
        }
        sb.append("\n");

        for (double hh : h) {
            sb.append(String.format("%.3f ", hh));
        }

        return sb.toString();
    }

    public void rndQuasideterm(double da, int ka, double delta) {
        if (da >= 0) rndQuasideterm(a, da, ka, true);
        for (double[] bb : b) {
            randomize3(bb, delta);
        }
    }

    public void rndQuasideterm(double da, int ka, double db, int kb, double dp, int kp) {
        if (da >= 0) rndQuasideterm(a, da, ka, true);
        if (db >= 0) rndQuasideterm(b, db, kb, false);
        if (dp >= 0) rndQuasideterm(new double[][]{h}, dp, kp, false);
        updateSampling();
        checkModel();
    }

    private void rndQuasideterm(double[] x, double d, int k, int p) {
        Arrays.fill(x, 0);
        if (k == -1) {
            k = x.length;
            d = 1;
        }
        x[p] = 1 - d;
        for (int i = 0; i < k; i++) {
            int ind = (int) (rnd.nextDouble() * x.length);
            x[ind] += d / k;
        }
    }

    private int rnd(List<Integer> a, int size) {
        int sz = a.isEmpty() ? size : a.size();
        return (int) (rnd.nextDouble() * sz);
    }

    private void rndQuasideterm(double[][] x, double d, int k, boolean conn) {
        for (int i = 0; i < x.length; i++) {
            int ind = (int) (rnd.nextDouble() * x[i].length);
            rndQuasideterm(x[i], d, k, conn ? (i + 1) % x.length : ind);
        }
    }

    public Hmm copy() {
        double na[][] = new double[n][n];
        double nb[][] = new double[n][m];
        double np[] = h.clone();
        for (int i = 0; i < n; i++) {
            na[i] = a[i].clone();
            nb[i] = b[i].clone();
        }
        return new Hmm(na, nb, np);
    }

    public void rndQuasidetermA(double delta, int pos[][]) {
        for (int i = 0; i < n; i++) {
            Arrays.fill(a[i], 0);
            a[i][(i + 1) % n] = 1 - delta;
            for (int j : pos[i]) {
                a[i][j] += delta / pos[i].length;
            }
        }
        checkModel();
        updateSampling();
    }

}
