package ru.ifmo.robocode;

import ru.ifmo.robocode.AutomataFunction;
import ru.ifmo.robocode.Environment;
import ru.ifmo.robocode.TankControlSystem;
import ru.ifmo.robocode.TankManager;
import ru.ifmo.robocode.ga.EvalChromosome;
import ru.ifmo.robocode.gep.Chromosome;
import robocode.AdvancedRobot;
import robocode.control.*;
import robocode.peer.RobotPeer;

import java.awt.geom.Rectangle2D;
import java.io.File;

public class Fitness implements ru.ifmo.robocode.ga.Fitness<Chromosome> {

    public static final Object mutex = new Object();

    private static final String ROBOTS_DIR = "robots";
    private static final int NUM_POS = 10;
    public static final int NUM_ROUNDS = NUM_POS * 2;

    private static final double GUN_COOLING_RATE = 0.1;

    private final double[][][] pos = new double[NUM_POS][2][3];
    private static final int BF_WIDTH = 800;
    private static final int BF_HEIGHT = 600;
    static final int INACTIVITY_TIME = 450;

    private static final double MIN_WEIGHT = 0.99;

    private double f;

    public Fitness() {
        pos[0] = new double[][]{{2 * RobotPeer.WIDTH, 2 * RobotPeer.HEIGHT, 0}, {BF_WIDTH - 2 * RobotPeer.WIDTH, BF_HEIGHT - 2 * RobotPeer.HEIGHT, 0}};
        pos[1] = new double[][]{{2 * RobotPeer.WIDTH, BF_HEIGHT - 2 * RobotPeer.HEIGHT, 0}, {BF_WIDTH - 2 * RobotPeer.WIDTH, 2 * RobotPeer.HEIGHT, 0}};
        for (int i = 0; i < NUM_POS - 2; i++) {
            pos[i + 2] = rndPos();
        }
        TankManager.getInstance().setPositions(pos);
    }

    public void fitness(EvalChromosome<Chromosome>[] pop) {
        double maxRes = 0;
        for (EvalChromosome<Chromosome> c : pop) {
            double min = Long.MAX_VALUE;
            double max = 0;
            final String[] enemies = TankManager.getInstance().getEnemies();
            for (String enemy : enemies) {
                double r = fitness(c.c, enemy);
                min = Math.min(min, r);
                max = Math.max(max, r);
            }
            c.v = min * MIN_WEIGHT + max * (1 - MIN_WEIGHT);
            maxRes = Math.max(maxRes, c.v);
        }
    }

    protected double[] rndPos(double other[][]) {
        Rectangle2D.Float r;
        double x, y;
        double h = 2 * Math.PI * Math.random();
        do {
            x = RobotPeer.WIDTH + Math.random() * (BF_WIDTH - 2 * RobotPeer.WIDTH);
            y = RobotPeer.HEIGHT + Math.random() * (BF_HEIGHT - 2 * RobotPeer.HEIGHT);
            r = getRect(x, y);
        } while (!ok(r, other));
        return new double[]{x, y, h};
    }

    private Rectangle2D.Float getRect(double x, double y) {
        return new Rectangle2D.Float((float) (x - RobotPeer.WIDTH / 2 + 2), (float) (y - RobotPeer.HEIGHT / 2 + 2),
                RobotPeer.WIDTH - 4, RobotPeer.HEIGHT - 4);
    }

    private boolean ok(Rectangle2D.Float r, double[][] other) {
        for (double[] v : other) {
            Rectangle2D.Float r2 = getRect(v[0], v[1]);
            if (r.intersects(r2)) {
                return false;
            }
        }
        return true;
    }

    private double[][] rndPos() {
        double p[][] = new double[2][3];
        p[0] = rndPos(new double[0][]);
        p[1] = rndPos(new double[][]{p[0]});
        return p;
    }

    private RobotSpecification getRobot(String clazz) {
        return RobotSpecFactory.create(new File(clazz + ".class"), new File(ROBOTS_DIR), "", false);
    }

    private boolean finished;

    private double fitness(Chromosome c, String enemyName) {
        TankManager.getInstance().setTank(createTank(c));
        RobocodeListener listener = new Listener();
        RobocodeEngine engine = new RobocodeEngine(listener);
        BattlefieldSpecification bfspec = new BattlefieldSpecification(BF_WIDTH, BF_HEIGHT);
        RobotSpecification enemy = getRobot(enemyName);
        RobotSpecification my = getRobot(TankManager.getInstance().getTank().getClass().getName());
        RobotSpecification[] rs = {enemy, my};
        BattleSpecification bspec = new BattleSpecification(NUM_ROUNDS, INACTIVITY_TIME, GUN_COOLING_RATE, bfspec, rs);
        finished = false;
        engine.runBattle(bspec);

        synchronized (mutex) {
            if (!finished) {
                try {
                    mutex.wait();
                } catch (InterruptedException e) {
                    throw new RuntimeException(e);
                }
            }
        }
        engine.close();
        return f;
    }

    private AdvancedRobot createTank(Chromosome c) {
        TankControlSystem function = new AutomataFunction(c, Basis.values(), Environment.terms());
        SimpleTank tank = new SimpleTank();
        tank.setModel(function);
        return tank;
    }

    private final class Listener implements RobocodeListener {
        public void battleComplete(BattleSpecification battle, RobotResults[] results, int numTiks, long ticksTime, long robotsTime) {
            SimpleTank tank = (SimpleTank) TankManager.getInstance().getTank();
            final String myClassName = tank.getClass().getName();
            for (int i = 0; i < 2; i++) {
                RobotResults res = results[i];
                final String className = res.getRobot().getClassName();
                if (!myClassName.equals(className)) {
                    continue;
                }

                int allScore = (res.getScore() + results[1 - i].getScore());
                double score = (allScore == 0) ? 0 : res.getScore() * 1.0 / allScore;

                f = score * 100;
            }
            synchronized (mutex) {
                finished = true;
                mutex.notifyAll();
            }
        }

        public void battleAborted(BattleSpecification battle) {
            System.out.println("Aborted!");
        }

        public void battleMessage(String message) {

        }

    }

}
